/**
 **	File ......... BlackAdder.cpp
 **	Published ....  2004-05-17
 **	Author ....... grymse@alhem.net
**/
/*
Copyright (C) 2004  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <string>
#include <vector>
#include <map>

#include "dp.h"
#include "BlackAdder.h"

//#define DEB(x) x; fflush(stdout);
#define DEB(x)


	typedef struct
	{
		std::string nsp;
		std::string cls;
	} ADD;

	static const char *colors[] = { "#ffffcc", "#ffccff", "#ccffff", "#ffcccc", "#ccffcc", "#ccccff", "#99ccff", "#cc99ff", "#99ffcc", "#ccff99" };
	static int colorcount = 0;
	static int maxcolor = 10;


BlackAdder::BlackAdder()
{
}


BlackAdder::~BlackAdder()
{
}


CLS *BlackAdder::add_class(const std::string& strNamespace,string_v& vecStr,const std::string& filename,int line,bool bExtern)
{
	NSP *nsp = GetNamespace(strNamespace);
	std::string name;
	if (0&&strNamespace.size())
		name = strNamespace + "::" + vecStr[1];
	else
		name = vecStr[1];
	for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
	{
		CLS *tmp = *it;
		if (!strcmp(tmp -> name.c_str(),name.c_str()))
			return tmp;
	}
	// class not added, so add
	CLS *tmp = new CLS;
	tmp -> name = name;
	tmp -> group_name = get_group(name);
	tmp -> m_namespace = strNamespace;
	tmp -> filename = filename;
	tmp -> line = line;
	tmp -> bExtern = bExtern;
	tmp -> bUsed = false;
	nsp -> added.push_back( tmp );
DEB(	fprintf(stderr,"Class to add to namespace '%s': %s\n",strNamespace.c_str(),name.c_str());)
	bool col = false;
	bool debug = false;
	int count = 0;
	std::string bas;
	if (strstr(name.c_str(),"Form") || strstr(name.c_str(),"Web") )
		debug = true;
	for (string_v::iterator it = vecStr.begin(); it != vecStr.end(); it++)
	{
		if (!col)
		{
			if (*it == ":")
				col = true;
		}
		else
		{
			if (*it == "<")
				count++;
			if (*it == ">")
				count--;
			if (*it == "{" || (*it == "," && !count))
			{
				tmp -> bases.push_back( bas );
				tmp -> bUsed = true;
				if (debug)
					fprintf(stderr," basklass: '%s'\n",bas.c_str());
				bas = "";
			}
			else
			{
				if (*it != "virtual" &&
				    *it != "public" &&
				    *it != "protected" &&
				    *it != "private")
				{
					if (bas.size())
						bas += " " + *it;
					else
					{
						bas = *it;
						for (strstr_m::iterator i2 = m_using.begin(); i2 != m_using.end(); i2++)
						{
							if (!strcmp((*i2).first.c_str(), bas.c_str()))
							{
								bas = (*i2).second;
								break;
							}
						}
					}
				}
			}
		}
	}
//	printf("\n");
	return tmp;
}


void BlackAdder::add_using(const std::string& str)
{
	size_t x = 0;
	for (size_t i = 0; i < str.size() - 1; i++)
		if (str[i] == ':' && str[i + 1] == ':')
			x = i + 2;
	m_using[str.c_str() + x] = str;
}


CLS *BlackAdder::find_class(const std::string& strNamespace,const std::string& strName)
{
	NSP *nsp = GetNamespace(strNamespace);
DEB(	printf("Looking for '%s :: %s'\n",strNamespace.c_str(),strName.c_str());)
	for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
	{
		CLS *cls = *it;
		if (!strcmp(cls -> name.c_str(),strName.c_str()))
			return cls;
	}
	return NULL;
}


void BlackAdder::clear_using()
{
	while (m_using.size())
		m_using.erase(m_using.begin());
}


void BlackAdder::add_type(CLS *cls,const std::string& str)
{
	// skip namespace resolution here
/*
	for (typedef_m::iterator it = m_typedef.begin(); it != m_typedef.end(); it++)
	{
		if (!strcmp((*it).first.c_str(),str.c_str()))
		{
			std::string prev;
//#define DEB(x) x
DEB(			printf("typedef '%s'\n",str.c_str());)
			for (string_v::iterator it2 = (*it).second.begin(); it2 != (*it).second.end(); it2++)
			{
				if (*it2 == "<")
				{
					if (prev.size())
						add_type(cls, prev);
					prev = "";
				}
				else
				if (*it2 == ",")
				{
					if (prev.size())
						add_type(cls, prev);
					prev = "";
				}
				else
				if (*it2 == ">")
				{
					if (prev.size())
						add_type(cls, prev);
					prev = "";
				}
				else
				{
//					if (prev.size())
//						prev += " ";
					prev += *it2;
				}
			}
DEB(			printf("typedef '%s' ok\n",str.c_str());)
			return;
		}
	}
*/
	for (string_v::iterator it = cls -> using_type.begin(); it != cls -> using_type.end(); it++)
	{
		std::string x = *it;
		if (!strcmp(x.c_str(), str.c_str()))
			return;
	}
fprintf(stderr, " adding type '%s' for class '%s'\n", str.c_str(), cls -> name.c_str());
	cls -> using_type.push_back( str );
}
//#define DEB(x)


NSP *BlackAdder::GetNamespace(const std::string& name,bool add)
{
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		if (!strcmp(nsp -> name.c_str(),name.c_str()))
		{
			return nsp;
		}
	}
	if (!add)
		return NULL;
	NSP *nsp = new NSP;
	nsp -> name = name;
	m_added.push_back(nsp);
	return nsp;
}


void BlackAdder::fix_class()
{
	std::vector<ADD *> c2add;
	return;

	// arv
	
	// namespace loop
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
DEB(		printf("\n fix_class() namespace '%s'\n",nsp -> name.c_str());)

		// class loop
		for (std::vector<CLS *>::iterator it2 = nsp -> added.begin(); it2 != nsp -> added.end(); it2++)
		{
			CLS *cls = *it2;
DEB(			printf(" fix_class()  class '%s'\n",cls -> name.c_str());)
			std::string name;
			if (nsp -> name.size())
				name = nsp -> name + "::" + cls -> name;
			else
				name = cls -> name;
DEB(			printf(" fix_class()  classname '%s'\n",name.c_str());)

			// base class loop
			for (string_v::iterator it3 = cls -> bases.begin(); it3 != cls -> bases.end(); it3++)
			{
				std::string base = *it3;
				bool resolved = false;

DEB(				printf(" fix_class() base_class '%s'\n",base.c_str());)

				// try to locate base class
				if (!strstr(base.c_str(),"::"))
				{
				for (std::vector<CLS *>::iterator i2 = nsp -> added.begin(); i2 != nsp -> added.end(); i2++)
				{
					CLS *cls = *i2;
					if (!strcmp(cls -> name.c_str(),base.c_str()))
					{
						if (nsp -> name.size())
							base = nsp -> name + "::" + base;
						resolved = true;
						cls -> bUsed = true;
						break;
					}
				}
				if (!resolved) // find class in blank namespace
				{
					for (std::vector<NSP *>::iterator i2 = m_added.begin(); i2 != m_added.end(); i2++)
					{
						NSP *nsp = *i2;
						if (!nsp -> name.size()) // blank namespace
						{
							for (std::vector<CLS *>::iterator i3 = nsp -> added.begin(); i3 != nsp -> added.end(); i3++)
							{
								CLS *cls = *i3;
								if (!strcmp(cls -> name.c_str(),base.c_str()))
								{
									resolved = true; // ok base class in blank namespace
									cls -> bUsed = true;
									break;
								}
							}
						}
					}
					if (!resolved) // base must be on our namespace, somehow
					{
						ADD *tmp = new ADD;
						tmp -> nsp = nsp -> name;
						tmp -> cls = base;
						c2add.push_back( tmp );

						if (nsp -> name.size())
							base = nsp -> name + "::" + base;
					}
				}
				} // !strstr(,"::")

				if (strcmp((*it3).c_str(),base.c_str()))
				{
DEB(					printf(" fix_class() renaming base class '%s' to '%s'\n",(*it3).c_str(),base.c_str());)
					(*it3) = base;
				}

				if (!resolved)
				{
					std::string strNamespace;
					std::string strName = base;
					for (size_t i = 0; i < base.size() - 1; i++)
					{
						if (base[i] == ':' && base[i + 1] == ':')
						{
							char slask[1000];
							strcpy(slask,base.c_str());
							slask[i] = 0;
							strNamespace = slask;
							strName = slask + i + 2;
						}
					}
					CLS *t2;
					if (!(t2 = find_class(strNamespace,strName)))
					{
						ADD *tmp = new ADD;
						tmp -> nsp = strNamespace;
						tmp -> cls = strName;
						c2add.push_back( tmp );
					}
					else
					{
DEB(						printf("setting bUsed for class '%s' in namespace '%s'\n",strName.c_str(),strNamespace.c_str());)
						t2 -> bUsed = true;
					}
				}
			}
		}
	}
DEB(	printf("fix_class() adding\n");)
	for (std::vector<ADD *>::iterator it = c2add.begin(); it != c2add.end(); it++)
	{
		ADD *tmp = *it;
		std::vector<std::string> vec;
		vec.push_back("class");
		vec.push_back(tmp -> cls);
		vec.push_back("{");
DEB(		printf("Adding '%s :: %s'\n",tmp -> nsp.c_str(),tmp -> cls.c_str());)
		CLS *t2 = add_class(tmp -> nsp,vec,"",0,true);
		t2 -> bUsed = true;
	}

	// relationer
DEB(	printf("fix_class() relationer\n");)
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
		{
			CLS *cls = *it;
			std::string name;
			if (nsp -> name.size())
				name = nsp -> name + "::" + cls -> name;
			else
				name = cls -> name;
			for (string_v::iterator it = cls -> using_type.begin(); it != cls -> using_type.end(); it++)
			{
				std::string typ = *it;
				char slask[1000];
				bool ptr = false;
				bool arr = false;
				bool ref = false;
				char *s;
				strcpy(slask,typ.c_str());
				if ((s = strstr(slask,"*")) != NULL)
				{
					ptr = true;
					*s = 0;
				}
				if ((s = strstr(slask,"[")) != NULL)
				{
					arr = true;
					*s = 0;
				}
				if ((s = strstr(slask,"&")) != NULL)
				{
					ref = true;
					*s = 0;
				}
				NSP *nsp2;
				CLS *cls2;
				if (locate(nsp,slask,nsp2,cls2))
				{
					cls -> bUsed = true;
					cls2 -> bUsed = true;
				}
			} // for (relationer...
		}
	}
	// /relationer
DEB(	printf("fix_class() ok\n");)
}


std::string darken(const std::string& color)
{
	std::string r = "#";
	for (int i = 1; i < 7; i++)
	{
		r += color[i];
		if (color[i] > '0' && color[i] < 'f')
		{
			if (color[i] == 'a')
				r[i] = '8';
			else
			if (color[i] == 'b')
				r[i] = '9';
			else
				r[i] -= 2;
		}
	}
	return r;
}


std::string merge_colors(const std::string&c1, const std::string& c2)
{
	std::string r;
	for (int i = 0; i < 7; i++)
		r += c1[i] < c2[i] ? c1[i] : c2[i];
	return r;
}


bool BlackAdder::has_subclass(NSP *nsp0, CLS *cls0, const std::string& color)
{
	bool r = false;
	std::string name0;
	if (nsp0 -> name.size())
		name0 = nsp0 -> name + "::" + cls0 -> name;
	else
		name0 = cls0 -> name;
DEB(fprintf(stderr, "Check subclass for '%s'\n", name0.c_str());)
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
		{
			CLS *cls = *it;
			std::string name;
			if (nsp -> name.size())
				name = nsp -> name + "::" + cls -> name;
			else
				name = cls -> name;
			for (string_v::iterator it = cls -> bases.begin(); it != cls -> bases.end(); it++)
			{
				std::string base = *it;
DEB(fprintf(stderr, " ... check base '%s'\n", base.c_str());)
				if (base == name0)
				{
					if (cls -> color.size())
					{
						cls -> color = darken(merge_colors(cls -> color, color));
					}
					else
					{
						cls -> color = darken(color);
					}
					r = true;
				}
/*
				if (base == cls0 -> name)
					return true;
*/
			}
		}
	}
DEB(fprintf(stderr, " has_subclass OK\n");)
	return r;
}


CLS *BlackAdder::find_class(const std::string& name0)
{
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
		{
			CLS *cls = *it;
			std::string name;
			if (nsp -> name.size())
				name = nsp -> name + "::" + cls -> name;
			else
				name = cls -> name;
			if (name0 == name)
				return cls;
			if (cls -> name == name0)
				return cls;
		}
	}
	return NULL;
}


std::list<CLS *> BlackAdder::find_base(NSP *nsp, CLS *cls)
{
	std::list<CLS *> bases;
	std::string name;
	bool debug = false;
	if (nsp -> name.size())
		name = nsp -> name + "::" + cls -> name;
	else
		name = cls -> name;
	for (string_v::iterator it = cls -> bases.begin(); it != cls -> bases.end(); it++)
	{
		std::string base = *it;
		CLS *base_cls = NULL;
		if (strstr(name.c_str(),"Form") ||
		    strstr(base.c_str(),"Form")
		   )
		{
			fprintf(stderr,"Class: '%s' : Base Class: '%s'\n",name.c_str(),base.c_str());
			debug = true;
		}
		base_cls = find_class(base);
DEB(fprintf(stderr, "Class '%s': %s\n", base.c_str(), base_cls ? "found" : "not found");)
/*
		for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
		{
			CLS *cls = *it;
			if (!strcmp(cls -> name.c_str(),base.c_str()))
			{
				if (!strstr(base.c_str(),"::") && nsp -> name.size())
					base = nsp -> name + "::" + base;
				resolved = true;
				base_cls = cls;
				if (debug)
					fprintf(stderr,"resolved as 1\n");
				break;
			}
		}
		if (!resolved && !strstr(base.c_str(),"::") ) // find class in blank namespace
		{
			for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
			{
				NSP *nsp = *it;
				if (!nsp -> name.size()) // blank namespace
				{
					for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
					{
						CLS *cls = *it;
						if (!strcmp(cls -> name.c_str(),base.c_str()))
						{
							resolved = true; // ok base class in blank namespace
							base_cls = cls;
							if (debug)
								fprintf(stderr,"resolved as 2\n");
							break;
						}
					}
				}
			}
			if (!resolved) // base must be on our namespace, somehow
			{
				if (!strstr(base.c_str(),"::") && nsp -> name.size())
				{
					base = nsp -> name + "::" + base;
					if (debug)
						fprintf(stderr,"not resolved - renamed: %s\n",base.c_str());
				}
			}
		}
*/
		if (base_cls)
		{
			bases.push_back(base_cls);
		}
	} // for (basklasser...
	return bases;
}


void BlackAdder::MakeDot(const std::string& group_name)
{
	bool bCluster = true;
DEB(	printf("MakeDot()\n");)	
	if (config["-cluster"] == "true")
		bCluster = false;

DEB(	printf("fix_class()\n");)
	fix_class();


//	---
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
		{
			CLS *cls = *it;
			if (cls -> group_name != group_name)
				continue;
			bool has_base = cls -> bases.size() ? true : false;
DEB(fprintf(stderr, "calling has_subclass\n");)
			if (!has_base && has_subclass(nsp, cls, colors[colorcount % maxcolor]))
			{
				cls -> color = colors[colorcount++ % maxcolor];
			}
		}
	}
//	---
DEB(	fprintf(stderr, "*** Assigning colors for subclasses\n");)
	for (int i = 0; i < 10; i++)
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
		{
			CLS *cls = *it;
			if (cls -> group_name != group_name)
				continue;
			std::list<CLS *> vec = find_base(nsp, cls);
			if (!cls -> color.size() || cls -> color == "#ffffff")
			{
				std::string color = "#ffffff";
				bool ok = true;
				for (std::list<CLS *>::iterator it = vec.begin(); it != vec.end(); it++)
				{
					CLS *cls_base = *it;
					if (cls_base -> color.size() == 0)
					{
						ok = false;
					}
					else
					{
						color = merge_colors(color, cls_base -> color);
					}
				}
				if (ok && color != "#ffffff")
				{
					cls -> color = darken(color);
				}
			}
		}
	}
//	---

	for (int empty = 1; empty >= 0; empty--)
	{
DEB(		printf(" empty = %d\n",empty);)
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		bool clust = false;
		if ((nsp -> name.size() && !empty) || (!nsp -> name.size() && empty))
		{
			if (nsp -> name.size() && (bCluster || box[nsp -> name.c_str()] == "true") )
			{
				printf("	subgraph cluster_%d {\n",cluster++);
				printf("		style=filled;\n");
				printf("		fillcolor=lightgrey;\n");
				printf("		label=\"Namespace '%s'\"\n",nsp -> name.c_str());
				printf("		node [shape=box style=filled fillcolor=white]\n");
				clust = true;
			}
			else
			{
//				printf("	node [shape=box]\n");
			}
			for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
			{
				CLS *cls = *it;
				if (cls -> group_name != group_name)
					continue;
//				bool has_base = cls -> bases.size() ? true : false;
//				find_color(cls);
				std::string name;
				std::string args; // [ ...args... ]
				if (nsp -> name.size())
					name = nsp -> name + "::" + cls -> name;
				else
					name = cls -> name;
				// build args
				if (cls -> filename.size())
				{
//					char slask[1000];
//					sprintf(slask," URL=\"index.html\" tooltip=\"%s(%d)\"",cls -> filename.c_str(),cls -> line);
//					args += slask;
					switch (cls -> filename[0])
					{
					case '/': // sys hdr (probably)
						args += " style=filled fillcolor=\"#e0a0a0\"";
						break;
					case '.':
						if (cls -> filename.size() > 1 && cls -> filename[1] == '.')
						{
							args += " style=filled fillcolor=\"#e0e0e0\"";
						}
						break;
					}
				}
				else
				if (cls -> bExtern)
				{
					args += " style=filled fillcolor=\"#e0a0a0\"";
				}
				if (!args.size())
				{
					char slask[1000];
					char tmp[1000];
					size_t j = 0;
					for (size_t i = 0; i < cls -> name.size(); i++)
					{
						if (cls -> name[i] == '_')
							tmp[j++] = '_';
						tmp[j++] = cls -> name[i];
					}
					tmp[j] = 0;
					sprintf(slask," URL=\"html/class%s.html\" tooltip=\"Class documentation\"",tmp);
					args += slask;
				}
				// add fontsize
				{
					char slask[100];
					sprintf(slask, " fontsize=%d", fontsize);
					args += slask;
				}
//				if (!has_base && has_subclass(nsp, cls) )
				if (cls -> color.size())
				{
					args += " style=filled fillcolor=\""; //#ffffe0\"";
					args += cls -> color; //colors[colorcount++ % maxcolor];
					args += "\"";
				}
				// ...
				{
					if (clust)
						printf("\t");
					if (args.size())
					{
						printf("\t\"%s\" [%s]\n",name.c_str(),args.c_str() + 1);
					}
					else
					{
						printf("\t\"%s\"\n",name.c_str());
					}
				}
			}
			if (clust)
			{
				printf("	}\n");
			}
		} // if (empty...
	}
	// arv
DEB(	printf("arv\n");)
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		bool debug = false;
		if ((nsp -> name.size() && !empty) || (!nsp -> name.size() && empty))
		{
		for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
		{
			CLS *cls = *it;
			if (cls -> group_name != group_name)
				continue;
			std::string name;
			if (nsp -> name.size())
				name = nsp -> name + "::" + cls -> name;
			else
				name = cls -> name;
			for (string_v::iterator it = cls -> bases.begin(); it != cls -> bases.end(); it++)
			{
				std::string base = *it;
				bool resolved = false;
				if (strstr(name.c_str(),"Form") ||
				    strstr(base.c_str(),"Form")
				   )
				{
					fprintf(stderr,"Class: '%s' : Base Class: '%s'\n",name.c_str(),base.c_str());
				   	debug = true;
				}
				for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
				{
					CLS *cls = *it;
					if (!strcmp(cls -> name.c_str(),base.c_str()))
					{
						if (!strstr(base.c_str(),"::") && nsp -> name.size())
							base = nsp -> name + "::" + base;
						resolved = true;
						if (debug)
							fprintf(stderr,"resolved as 1\n");
						break;
					}
				}
				if (!resolved && !strstr(base.c_str(),"::") ) // find class in blank namespace
				{
					for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
					{
						NSP *nsp = *it;
						if (!nsp -> name.size()) // blank namespace
						{
							for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
							{
								CLS *cls = *it;
								if (!strcmp(cls -> name.c_str(),base.c_str()))
								{
									resolved = true; // ok base class in blank namespace
									if (debug)
										fprintf(stderr,"resolved as 2\n");
									break;
								}
							}
						}
					}
					if (!resolved) // base must be on our namespace, somehow
					{
						if (!strstr(base.c_str(),"::") && nsp -> name.size())
						{
							base = nsp -> name + "::" + base;
							if (debug)
								fprintf(stderr,"not resolved - renamed: %s\n",base.c_str());
						}
					}
				}
				printf("	\"%s\" -> \"%s\" [dir=back arrowtail=empty]\n",
					base.c_str(),name.c_str());
			} // for (basklasser...
		}
		} // if (empty...
	}
	} // for (empty)

	// relationer
DEB(	printf("relationer\n");)
	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
		{
			CLS *cls = *it;
			if (cls -> group_name != group_name)
				continue;
			std::string name;
			if (nsp -> name.size())
				name = nsp -> name + "::" + cls -> name;
			else
				name = cls -> name;
			for (string_v::iterator it = cls -> using_type.begin(); it != cls -> using_type.end(); it++)
			{
				std::string typ = *it;
				char slask[1000];
				bool ptr = false;
				bool arr = false;
				bool ref = false;
				char *s;
				strcpy(slask,typ.c_str());
				if ((s = strstr(slask,"*")) != NULL)
				{
					ptr = true;
					*s = 0;
				}
				if ((s = strstr(slask,"[")) != NULL)
				{
					arr = true;
					*s = 0;
				}
				if ((s = strstr(slask,"&")) != NULL)
				{
					ref = true;
					*s = 0;
				}
				NSP *nsp2;
				CLS *cls2;
				if (strncmp(slask, "int", 3) && locate(nsp,slask,nsp2,cls2,false))
				{
					std::string name2;
					if (nsp2 -> name.size())
						name2 = nsp2 -> name + "::" + cls2 -> name;
					else
						name2 = cls2 -> name;
//					printf("	\"%s\" [fontsize=%d]\n", name.c_str(), fontsize);
					if (ptr)
					{
						printf("	\"%s\" -> \"%s\" [arrowtail=none arrowhead=open style=dashed]\n",
							name.c_str(),
							name2.c_str());
					}
					else
					if (ref)
					{
						printf("	\"%s\" -> \"%s\" [arrowtail=odiamond arrowhead=none style=dashed]\n",
							name.c_str(),
							name2.c_str());
					}
					else
					{
						printf("	\"%s\" -> \"%s\" [arrowtail=diamond arrowhead=none style=dashed]\n",
							name.c_str(),
							name2.c_str());
					}
				}
				else
				{
				}
			} // for (relationer...
		}
	}
	// /relationer
DEB(	printf("MakeDot() ok\n");)
}


bool BlackAdder::locate(NSP *base,const char *name,NSP *&nsp2,CLS *&cls2,bool add)
{
	std::string strUsing;
	bool found = false;
	for (strstr_m::iterator it = m_using.begin(); it != m_using.end(); it++)
	{
		if (!strcmp((*it).first.c_str(), name))
		{
			strUsing = (*it).second;
			found = true;
			break;
		}
	}
	if (found && strUsing.size())
	{
		size_t x = 0;
		char slask[3000];
		for (size_t i = 0; i < strUsing.size() - 2; i++)
			if (strUsing[i] == ':' && strUsing[i + 1] == ':')
				x = i;
		strcpy(slask,strUsing.c_str());
		slask[x] = 0;
		nsp2 = GetNamespace(slask,add);
		for (std::vector<CLS *>::iterator it = nsp2 -> added.begin(); it != nsp2 -> added.end(); it++)
		{
			CLS *cls = *it;
			if (!strcmp(cls -> name.c_str(),name))
			{
				cls2 = cls;
				return true;
			}
		}
		return false; // huh
	}
	else
	{
		nsp2 = base;
		for (std::vector<CLS *>::iterator it = nsp2 -> added.begin(); it != nsp2 -> added.end(); it++)
		{
			CLS *cls = *it;
			if (!strcmp(cls -> name.c_str(),name))
			{
				cls2 = cls;
				return true;
			}
		}
		nsp2 = GetNamespace("",add);
		if (!nsp2)
		{
			fprintf(stderr,"Couldn't find empty namespace\n");
		}
		else
		for (std::vector<CLS *>::iterator it = nsp2 -> added.begin(); it != nsp2 -> added.end(); it++)
		{
			CLS *cls = *it;
			if (!cls)
				fprintf(stderr,"locate, cls is NULL\n");
			else
			{
				if (!strcmp(cls -> name.c_str(),name))
				{
					cls2 = cls;
					return true;
				}
			}
		}
	}
	return false;

	for (std::vector<NSP *>::iterator it = m_added.begin(); it != m_added.end(); it++)
	{
		NSP *nsp = *it;
		for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
		{
			CLS *cls = *it;
			if (!strcmp(cls -> name.c_str(),name))
			{
				nsp2 = nsp;
				cls2 = cls;
				return true;
			}
		}
	}
	return false;
}


void BlackAdder::set_typedef(const std::string& nspace,const std::string& t,std::vector<std::string>&def)
{
	NSP *nsp = GetNamespace(nspace);
	nsp -> typedefs[t] = def;
}


bool BlackAdder::type_lookup(CLS *cls,const std::string& typ,string_v& translated,NSP *& nsp)
{
	nsp = GetNamespace(cls -> m_namespace);
	for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
	{
		CLS *cls = *it;
		if (!strcmp(cls -> name.c_str(),typ.c_str()))
		{
			return true;
		}
	}
	for (typedef_m::iterator it = nsp -> typedefs.begin(); it != nsp -> typedefs.end(); it++)
	{
		std::string typname = (*it).first;
		string_v& ref = (*it).second;
		if (!strcmp(typname.c_str(), typ.c_str()))
		{
			translated = ref;
			return true;
		}
	}
	nsp = GetNamespace("", false);
	if (!nsp)
		return false;
	for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
	{
		CLS *cls = *it;
		if (!strcmp(cls -> name.c_str(),typ.c_str()))
		{
			return true;
		}
	}
	for (typedef_m::iterator it = nsp -> typedefs.begin(); it != nsp -> typedefs.end(); it++)
	{
		std::string typname = (*it).first;
		string_v& ref = (*it).second;
		if (!strcmp(typname.c_str(), typ.c_str()))
		{
			translated = ref;
			return true;
		}
	}
	return false;
}


bool BlackAdder::is_class(const std::string& nspace,const std::string& name)
{
	NSP *nsp = GetNamespace(name, false);
	if (!nsp)
		return false;
	for (std::vector<CLS *>::iterator it = nsp -> added.begin(); it != nsp -> added.end(); it++)
	{
		CLS *cls = *it;
		if (!strcmp(cls -> name.c_str(), name.c_str()))
		{
			return true;
		}
	}
	return false;
}


