/**
 **	File ......... Scope.cpp
 **	Published ....  2004-05-17
 **	Author ....... grymse@alhem.net
**/
/*
Copyright (C) 2004  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
//#include <stdio.h>
#include <string>
#include <vector>
#include <map>

#include "Token.h"
#include "Scope.h"
#include "dp.h"
#include "SimpleParser.h"



Scope::Scope()
:Sequence()
{
}


Scope::Scope(token_v &tok)
:Sequence(tok)
{
}


Scope::~Scope()
{
	while (m_sts.size())
	{
		sequence_v::iterator it = m_sts.begin();
		Sequence *s = *it;
		delete s;
		m_sts.erase(it);
	}
}


void Scope::Display()
{
	if (IsExtern())
		return;
	if (config["-xml"] == "true")
	{
		if (Trail())
		{
			printf("<scope filename=\"%s\" line=\"%d\" value=\"%s\" trail=\"%s\">\n",
				GetFile().c_str(),
				GetLine(),
				GetExpr().c_str(),
				GetTrail().c_str());
		}
		else
			printf("<scope filename=\"%s\" line=\"%d\" value=\"%s\">\n",
				GetFile().c_str(),
				GetLine(),
				GetExpr().c_str());
		for (sequence_v::iterator it = m_sts.begin(); it != m_sts.end(); it++)
		{
			Sequence *s = *it;
			s -> Display();
		}
		printf("</scope>\n");
		return;
	}
	if (config["-dot"] == "true")
		return;
	if (GetParent()) // root Scope has parent NULL
	{
		Tab();
		printf("%s\n",GetExpr().c_str());
		Tab();
		printf("{\n");
	}
	for (sequence_v::iterator it = m_sts.begin(); it != m_sts.end(); it++)
	{
		Sequence *s = *it;
		s -> Display();
	}
	if (GetParent())
	{
		Tab();
		printf("}");
		if (Trail())
		{
			printf(" %s",GetTrail().c_str());
/*
			for (token_v::iterator it = m_trail.begin(); it != m_trail.end(); it++)
			{
				Token *t = *it;
				printf(" %s",t -> GetVal().c_str());
			}
//			printf(";");
*/
		}
		printf(" // %s\n",GetExpr().c_str());
	}
}


bool Scope::Trail()
{
	Token *t = GetFirstToken();
	if (!t)
		return false;
	switch (t -> GetCode())
	{
	case Typedef:
	case Class:
	case Struct:
	case Enum:
		return true;
	default:
		break;
	}
	return false;
}


void Scope::AddTrail(token_v &tok)
{
	m_trail = tok;
}


std::string Scope::GetTrail()
{
	std::string tmp;
	for (token_v::iterator it = m_trail.begin(); it != m_trail.end(); it++)
	{
		Token *t = *it;
		std::string str = t -> GetVal();
//		if (tmp.size() && (isalpha(str[0]) || str[0] == '<') )
//			tmp += " ";
		tmp += str;
	}
	if (config["-xml"] == "true")
		xmlsafestr(tmp);
	return tmp;
}


