/**
 **	File ......... SimpleParser.cpp
 **	Published ....  2004-05-17
 **	Author ....... grymse@alhem.net
**/
/*
Copyright (C) 2004  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <stdio.h>
#include <string>
#include <vector>
#include <map>
#include "Parse.h"
#include <assert.h>
#include <ctype.h>
#include "BlackAdder.h"
//#include "Statement.h"
#include "SimpleParser.h"
#include "dp.h"

#define DEB(x)



SimpleParser::SimpleParser(const std::string &path,BlackAdder *pAdder)
:BaseParser()
,the_ptr(0)
,m_state(0)
,m_currentfile("")
,m_path(path)
,m_line(0)
,m_namespace("")
,m_current(0)
,m_pAdder(pAdder)
,m_plevelm(0)
,m_stfile("")
,m_stline(0)
,m_extern(false)
{
	*the_str = 0;
}


SimpleParser::~SimpleParser()
{
}


void SimpleParser::error(const std::string &str)
{
DEB(	fprintf(stderr,"%s line %d: %s\n",m_currentfile.c_str(),m_line,str.c_str());)
}


void SimpleParser::include_file(const std::string &filename)
{
	char tmp[4000];

	strcpy(tmp,filename.c_str() + 1);
	while (*tmp == ' ' || *tmp == '\t')
		memmove(tmp,tmp + 1,strlen(tmp + 1));
	for (size_t i = 0; i < strlen(tmp); i++)
		if (tmp[i] == '"' || tmp[i] == '>')
			tmp[i] = 0;

//	printf("Including file '%s'...\n",filename.c_str());

	if (filename[0] == 34)
	{
		FILE *fil;
		char tmp2[4000];
		
		sprintf(tmp2,"%s/%s",m_path.c_str(),tmp);
		if ((fil = fopen(tmp2,"rt")) != NULL)
		{
			int line = 0;
			Parser(fil,line,tmp);
			fclose(fil);
		}
		else
		{
			sprintf(tmp2,"Couldn't open '%s'...",tmp);
			error( tmp2 );
		}
	}
	else
	{
		FILE *fil;
		char tmp2[4000];

		sprintf(tmp2,"/usr/include/%s",tmp);
		fil = fopen(tmp2,"rt");
		if (!fil)
		{
			sprintf(tmp2,"/usr/include/linux/%s",tmp);
			fil = fopen(tmp2,"rt");
		}
		if (!fil)
		{
			sprintf(tmp2,"/usr/include/g++-3/%s",tmp);
			fil = fopen(tmp2,"rt");
		}
		if (!fil)
		{
			sprintf(tmp2,"/usr/local/include/%s",tmp);
			fil = fopen(tmp2,"rt");
		}
		if (!fil)
		{
			sprintf(tmp2,"/usr/devel/include/%s",tmp);
			fil = fopen(tmp2,"rt");
		}
		if (!fil)
		{
			sprintf(tmp2,"%s/../%s",m_path.c_str(),tmp);
			fil = fopen(tmp2,"rt");
		}
		if (fil)
		{
			int line = 0;
			Parser(fil,line,tmp2);
			fclose(fil);
		}
		else
		{
			sprintf(tmp2,"Couldn't open '%s'...",tmp);
			error( tmp2 );
		}
	}
}


void SimpleParser::preprocessor(std::string &val,int &line)
{
	if (!strncmp(val.c_str(),"#include ",9))
	{
		include_file(val.c_str() + 9);
	}
	else
	{
		Parse pa( (char *)val.c_str() );
		char slask[200];

		strcpy(slask,pa.getword().c_str()); // '#'
		assert(*slask == '#');
		strcpy(slask,pa.getword().c_str()); // line
		line = atoi(slask);
		m_line = line;
		strcpy(slask,pa.getword().c_str());
		if (*slask == 34)
		{
			memmove(slask,slask + 1,strlen(slask + 1) + 1);
			slask[strlen(slask) - 1] = 0;
		}
		m_currentfile = slask;
		if (*slask == '/')
			m_extern = true;
		else
			m_extern = false;
	}
}


int SimpleParser::ReadLine(FILE *fil,int &line)
{
	fgets(the_str,5000,fil);
	line++;
	m_line = line;
	if (feof(fil))
		return 0;
	while (strlen(the_str) && (Cx == 13 || Cx == 10))
		Cx = 0;
	while (Cx == '\\') // continue
	{
		fgets(the_str + strlen(the_str) - 1,5000 - strlen(the_str),fil);
		line++;
		m_line = line;
		if (feof(fil))
			return 0;
		while (strlen(the_str) && (Cx == 13 || Cx == 10))
			Cx = 0;
	}
	the_ptr = 0;
	return 1;
}


int SimpleParser::GetToken(FILE *fil,int &line,std::string &val)
{
	while (C == ' ' || C == '\t')
		the_ptr++;
	while (!C)
	{
		if (!ReadLine(fil,line))
			return 0;
		if (C == '#')
		{
			val = the_str;
			C = 0;
			return CPP;
		}
		while (C == ' ' || C == '\t')
			the_ptr++;
	}
	if (isdigit(C))
	{
		while (isdigit(C) || C == '.')
		{
			val += C;
			the_ptr++;
		}
		return NUMBER;
	}
	if (isalpha(C) || C == '_' || C == '~' || (C == ':' && C1 == ':' && (isalpha(C2) || C2 == '_')) )
	{
		if (C == '~')
		{
			val += C;
			the_ptr++;
		}
		while (isalnum(C) || C == '_' || (C == ':' && C1 == ':'))
		{
			if (C == ':' && C1 == ':')
			{
				val += C;
				the_ptr++;
			}
			val += C;
			the_ptr++;
		}
		{
			int id = GetWord(val);
			if (id >= 256)
			{
				return id;
			}
		}
		return IDENTIFIER;
	}
	if (C == '/' && C1 == '*')
	{
		while (C != '*' || C1 != '/')
		{
			val += C;
			the_ptr++;
			while (!C)
			{
				if (!ReadLine(fil,line))
					return 0;
				val += '\n';
			}
		}
		val += C;
		the_ptr++;
		val += C;
		the_ptr++;
		return COMMENT;
	}
	if (C == '/' && C1 == '/')
	{
		val = the_str + the_ptr;
		C = 0;
		return COMMENT;
	}
	if (C == 34)
	{
		val += C;
		the_ptr++;
		while (C != 34 && C)
		{
			if (C == '\\') // escaping
			{
				val += C;
				the_ptr++;
			}
			val += C;
			the_ptr++;
		}
		if (C == 34)
		{
			val += C;
			the_ptr++;
		}
		else
		{
			error("Unterminated string constant");
		}
		return STRING;
	}
	if (C == '\'')
	{
		val += C;
		the_ptr++;
		while (C != '\'' && C)
		{
			if (C == '\\') // escaping
			{
				val += C;
				the_ptr++;
			}
			val += C;
			the_ptr++;
		}
		if (C == '\'')
		{
			val += C;
			the_ptr++;
		}
		else
		{
			error("Unterminated char constant");
		}
		return CHAR;
	}
	// dbl chars
#define dblcheck(x1,x2,y)  \
	if (C == x1 && C1 == x2) \
	{ \
		val = x1; \
		val += x2; \
		the_ptr += 2; \
		return y; \
	} 
	//
	dblcheck(':',':',CC);
	dblcheck('<','<',LTLT);
	dblcheck('>','>',GTGT);
	dblcheck('*','*',DPTR);
	dblcheck('<','=',LTEQ);
	dblcheck('!','=',NEQ);
	dblcheck('>','=',GTEQ);
	dblcheck('=','=',EQ);
	dblcheck('|','|',OR);
	dblcheck('&','&',AND);
	dblcheck('+','+',PP);
	dblcheck('-','-',MM);
	dblcheck('-','>',POINTER);
	// single char:
	val += C;
	if (C == '{')
	{
		m_namesave[m_level] = m_namespace;
		m_level++;
	}
	if (C == '(')
		m_plevel++;
	return the_str[the_ptr++];
}


void SimpleParser::Parser(FILE *fil,int &line,const std::string &name)
{
	CLS *cls = NULL;
	int c;
	std::string val;
	std::string old = m_currentfile;
	std::string strClass;
	bool func = false;
	bool ptr = false;
	std::string prev;

	m_currentfile = name;

	if (!old.size())
	{
//		printf("Parsing '%s'...\n",name.c_str());
	}

	for (;;)
	{
		val = "";
		c = GetToken(fil,line,val);
		if (!c) // eof
		{
			break;
		}
		if (c == IDENTIFIER && strstr(val.c_str(), "::"))
		{
DEB(			fprintf(stderr, "%s\n", val.c_str());)
		}
		if (c == CPP)
		{
			preprocessor( val, line );
			m_line = line;
		}
		else
		// if c is '{', m_level has already been increased
		// if c is '}', m_level is decreased after this switch
		switch (m_state)
		{
		case 0:
			if (m_vec.size())
			{
				while (m_vec.size())
					m_vec.erase(m_vec.begin());
			}
			switch (c)
			{
			case Class:
				m_state = 1;
				break;
			case Namespace:
				m_state = 2;
				break;
			case Using:
				m_state = 3;
				break;
			case Template:
				m_state = 4;
				break;
			case Friend:
				m_state = 5;
				break;
			case Typedef:
				m_state = 6;
				m_levelm = m_level;
				break;
			}
			if (m_state)
			{
				m_vec.push_back( val );
			}
			break;
		case 1: // class
			if (c != COMMENT)
				m_vec.push_back(val);
			if (c == ';') // forward decl
			{
				m_state = 0;
			}
			else
			if (c == '{') // definition
			{
				// uses m_namespace, m_vec
				cls = m_pAdder -> add_class(m_namespace,m_vec,m_stfile,m_stline,m_extern);
				if (m_namespace.size())
					m_namespace += "::" + m_vec[1];
				else
					m_namespace = m_vec[1];
				m_current = Class;
				m_state = 10;
				m_levelm = m_level;
				strClass = m_vec[1];
				// init class parse
				while (m_vec.size())
					m_vec.erase(m_vec.begin());
				func = false;
				ptr = false;
				// ...
			}
			break;
		case 2: // namespace
			if (c != COMMENT)
				m_vec.push_back(val);
			if (c == ';') // forward decl 'namespace ugly int beep();' ???
				m_state = 0;
			else
			if (c == '{') // definition
				m_state = 0;
			if (!m_state && m_vec[1] != "{")
			{
				if (m_namespace.size())
					m_namespace += "::" + m_vec[1];
				else
					m_namespace = m_vec[1];
				m_current = Namespace;
			}
			break;
		case 3: // using
			if (c != COMMENT)
				m_vec.push_back(val);
			if (c == ';') // forward decl 'namespace ugly int beep();' ???
				m_state = 0;
			else
			if (c == '{') // definition
				m_state = 0;
			if (!m_state)
			{
				m_pAdder -> add_using(m_vec[1]);
			}
			break;
		case 4: // template
			if (c != COMMENT)
				m_vec.push_back(val);
			if (c == ';') // forward decl 'namespace ugly int beep();' ???
				m_state = 0;
			else
			if (c == '{') // definition
				m_state = 0;
			break;
		case 5: // friend
			if (c != COMMENT)
				m_vec.push_back(val);
			if (c == ';') // forward decl 'namespace ugly int beep();' ???
				m_state = 0;
			else
			if (c == '{') // definition
				m_state = 0;
			break;
		case 6: // typedef
			if (c != COMMENT)
				m_vec.push_back(val);
			if (c == ';' && m_level == m_levelm)
				m_state = 0;
			if (!m_state)
			{
				std::string prev;
				std::string prev2;
				string_v::iterator it = m_vec.begin();
				string_v def;
				while (it != m_vec.end())
				{
					std::string tmp = *it;
					it++;
					if (prev2.size() && prev2 != "typedef")
					{
						def.push_back(prev2);
					}
					prev2 = prev;
					if (it != m_vec.end())
					{
						prev = tmp;
					}
				}
				m_pAdder -> set_typedef(m_namespace, prev, def);
			}
			break;
		case 10: // in class
			if (m_level < m_levelm)
				m_state = 0;
			else
			if (m_level == m_levelm) // not in inline method
			{
				if (c != COMMENT)
					m_vec.push_back(val);
				switch (val[0])
				{
				case ';':
					if (!func)
					{
						if (0)
						{
							printf("HEADER");
							for (size_t i = 0; i < m_vec.size(); i++)
								printf(" %s", m_vec[i].c_str());
							printf("\n");
						}
						if (m_vec[0] != "typedef" && m_vec[0] != "friend" )
						{
							std::string prev;
							std::string typ;
							find_types(cls, m_vec);
/*
							for (string_v::iterator it = m_vec.begin(); it != m_vec.end(); it++)
							{
								string_v translated;
								NSP *nsp;
								bool is_type = m_pAdder -> type_lookup(cls, *it, translated, nsp);
								if (is_type)
								{
									if (translated.size())
									{
										printf("TYPE*<%s, %s>", nsp -> name.c_str(), (*it).c_str());
										for (size_t i = 0; i < m_vec.size(); i++)
											printf(" %s", m_vec[i].c_str());
										printf("\n");
									}
									else
									{
										printf("TYPE<%s> %s\n", nsp -> name.c_str(), (*it).c_str());
									}
								}
								if (prev.size() && *it != "[" && (*it != ";" || prev == "]") &&
									prev != "static" && prev != "const" && *it != "=" && !eq
								)
								{
									if (prev == "<") // template 
										typ += " " + prev;
									else
										typ += prev;
								}
								if (*it == "=")
									eq = true;
								if (*it == ";")
									break;
								prev = *it;
							}
							if (cls)
							{
//printf("Calling add_type '%s' class '%s'\n", typ.c_str(), cls -> name.c_str());
								m_pAdder -> add_type(cls, typ);
							}
*/
						}
					}
					// init class parse
					while (m_vec.size())
						m_vec.erase(m_vec.begin());
					func = false;
					ptr = false;
					// ...
					break;
				case ':':
					if (m_vec[0] == "public")
					{
					}
					if (m_vec[0] == "protected")
					{
					}
					if (m_vec[0] == "private")
					{
					}
					// init class parse
					while (m_vec.size())
						m_vec.erase(m_vec.begin());
					func = false;
					ptr = false;
					// ...
					break;
				case '(':
					func = true;
					break;
				case '*':
					if (!func)
						ptr = true;
					break;
				default:
					break;
				}
			}
			break;
		}
		//
		if (c == '}')
		{
			m_level--;
			m_namespace = m_namesave[m_level];
		}
		else
		{
			EndToken(c);
		}
		prev = val;
	}
	m_currentfile = old;
}


void SimpleParser::find_types(CLS *cls,string_v& vec,int l)
{
	std::string typ;
	for (string_v::iterator it = vec.begin(); it != vec.end(); it++)
	{
		string_v translated;
		NSP *nsp;
		bool is_type = m_pAdder -> type_lookup(cls, *it, translated, nsp);
		if (*it == "&")
			typ += *it;
		else
		if (*it == "*")
			typ += *it;
		else
		if (*it == "[")
			typ += *it;
		else
		if (is_type)
		{
			if (translated.size())
			{
DEB(				fprintf(stderr,"TYPE*<%s, %s>", nsp -> name.c_str(), (*it).c_str());
				for (size_t i = 0; i < translated.size(); i++)
					fprintf(stderr," %s", translated[i].c_str());
				fprintf(stderr,"\n");)
				if (l < 5)
					find_types(cls, translated, l + 1);
			}
			else
			{
DEB(				fprintf(stderr,"TYPE<%s> %s\n", nsp -> name.c_str(), (*it).c_str());)
				if (typ.size())
					m_pAdder -> add_type(cls, typ);
				typ = *it;
			}
		}
	}
	if (typ.size())
		m_pAdder -> add_type(cls, typ);
}


